import contextlib
from dataclasses import dataclass
from urllib.parse import urlparse

from returns.functions import raise_exception
from returns.future import FutureResultE, future_safe
from returns.pipeline import flow
from returns.pointfree import bind

from abuse_whois import errors, schemas, settings
from abuse_whois.matchers.shared_hosting import get_shared_hosting_provider
from abuse_whois.matchers.whois import get_whois_contact
from abuse_whois.utils import (
    get_registered_domain,
    is_domain,
    is_email,
    is_ipv4,
    is_ipv6,
    is_url,
    resolve,
)
from abuse_whois.whois import query

from .abstract import AbstractService


@dataclass
class Container:
    address: str
    hostname: str


@dataclass
class ContainerWithRecords(Container):
    ip_address: str | None
    domain_record: schemas.WhoisRecord | None
    ip_record: schemas.WhoisRecord | None


@future_safe
async def validate_address(address: str) -> str:
    funcs = [is_email, is_domain, is_ipv4, is_ipv6, is_url]
    for f in funcs:
        if f(address):
            return address

    raise errors.AddressError(f"{address} is not supported type address")


@future_safe
async def get_hostname(address: str) -> Container:
    if is_ipv6(address) or is_ipv4(address) or is_domain(address):
        return Container(hostname=address, address=address)

    url_or_email = address
    if is_email(url_or_email):
        url_or_email = f"http://{address}"

    parsed = urlparse(url_or_email)
    if parsed.hostname is None:
        raise errors.AddressError(f"{address} does not have hostname")

    return Container(hostname=parsed.hostname, address=address)


@future_safe
async def whois_query(
    hostname, *, timeout: int = settings.QUERY_TIMEOUT
) -> schemas.WhoisRecord:
    return await query(hostname, timeout=timeout)


async def get_domain_records(
    container: Container, *, timeout: int = settings.QUERY_TIMEOUT
) -> ContainerWithRecords:
    domain_record = (
        (await whois_query(container.hostname, timeout=timeout).awaitable())
        .alt(raise_exception)
        .unwrap()
        ._inner_value
    )

    # get IP address by domain
    ip_record: schemas.WhoisRecord | None = None
    ip_address: str | None = None
    with contextlib.suppress(OSError):
        ip_address = await resolve(container.hostname)

    if ip_address is not None:
        ip_record = (
            await whois_query(container.hostname, timeout=timeout).awaitable()
        )._inner_value.value_or(None)

    return ContainerWithRecords(
        address=container.address,
        hostname=container.hostname,
        domain_record=domain_record,
        ip_record=ip_record,
        ip_address=ip_address,
    )


async def get_ip_records(
    container: Container, *, timeout: int = settings.QUERY_TIMEOUT
) -> ContainerWithRecords:
    ip_record = (
        (await whois_query(container.hostname, timeout=timeout).awaitable())
        .alt(raise_exception)
        .unwrap()
        ._inner_value
    )

    return ContainerWithRecords(
        address=container.address,
        hostname=container.hostname,
        domain_record=None,
        ip_record=ip_record,
        ip_address=container.hostname,
    )


@future_safe
async def get_records(
    container: Container, *, timeout: int = settings.QUERY_TIMEOUT
) -> ContainerWithRecords:
    if is_domain(container.hostname):
        return await get_domain_records(container, timeout=timeout)

    return await get_ip_records(container, timeout=timeout)


@future_safe
async def get_contacts(container: ContainerWithRecords):
    main_record = (
        container.domain_record
        if is_domain(container.hostname)
        else container.ip_record
    )
    if main_record is None:
        raise errors.NotFoundError(
            f"Record for {container.hostname} is not found or something went wrong"
        )

    registered_domain: str | None = None
    if is_domain(container.hostname):
        # set registered domain
        registered_domain = get_registered_domain(container.hostname)

    registrar: schemas.Contact | None = None
    if container.domain_record is not None:
        registrar = get_whois_contact(container.domain_record)

    hosting_provider: schemas.Contact | None = None
    if container.ip_record is not None:
        hosting_provider = get_whois_contact(container.ip_record)

    shared_hosting_provider = get_shared_hosting_provider(container.hostname)

    return schemas.Contacts(
        address=container.address,
        hostname=container.hostname,
        ip_address=container.ip_address,
        registered_domain=registered_domain,
        shared_hosting_provider=shared_hosting_provider,
        registrar=registrar,
        hosting_provider=hosting_provider,
        record=main_record,
    )


class ContactsQuery(AbstractService):
    async def call(self, address: str) -> schemas.Contacts:
        result: FutureResultE[schemas.Contacts] = flow(
            address,
            validate_address,
            bind(get_hostname),
            bind(get_records),
            bind(get_contacts),
        )
        return (await result.awaitable()).alt(raise_exception).unwrap()._inner_value



import asyncio
import json

import typer
from asyncwhois.errors import WhoIsError
from whodap.errors import WhodapError

from . import errors, schemas, services

app = typer.Typer()


async def get_abuse_contacts(address: str) -> schemas.Contacts:
    try:
        return await services.ContactsQuery().call(address)
    except (WhodapError, WhoIsError) as e:
        raise errors.AbuseWhoisError(str(e)) from e


@app.command()
def whois(
    address: str = typer.Argument(..., help="URL, domain, IP address or email address")
):
    try:
        contacts = asyncio.run(get_abuse_contacts(address))
        print(contacts.model_dump_json(by_alias=True))  # noqa: T201
    except (errors.AbuseWhoisError, asyncio.TimeoutError, ConnectionResetError) as e:
        print(json.dumps({"error": str(e)}))  # noqa: T201


if __name__ == "__main__":
    app()



import asyncio

from fastapi import APIRouter, HTTPException, status

from abuse_whois import errors, get_abuse_contacts, schemas

router = APIRouter(prefix="/whois")


@router.post("/", response_model=schemas.Contacts)
async def whois(query: schemas.Query) -> schemas.Contacts:
    try:
        return await get_abuse_contacts(query.address)
    except errors.AddressError as e:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail=str(e)) from e
    except asyncio.TimeoutError as e:
        raise HTTPException(status.HTTP_408_REQUEST_TIMEOUT, detail=str(e)) from e
    except errors.RateLimitError as e:
        raise HTTPException(status.HTTP_429_TOO_MANY_REQUESTS, detail=str(e)) from e
    except errors.NotFoundError as e:
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail=str(e)) from e
    except errors.AbuseWhoisError as e:
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)) from e
    except ConnectionResetError as e:
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)) from e



from abuse_whois import schemas


class WhoisRule(schemas.BaseRule):
    def match(self, record: schemas.WhoisRecord) -> bool:
        data = record.model_dump(by_alias=True)
        return super().match(data)



import asyncio
import pathlib
import socket
from collections.abc import Callable
from contextlib import contextmanager
from typing import cast

import tldextract
import validators
import yaml
from asyncache import cached
from asyncer import asyncify
from cachetools import TTLCache
from starlette.datastructures import CommaSeparatedStrings

from . import settings


def _is_x(v: str, *, validator: Callable[[str], bool]) -> bool:
    res = validator(v)

    if isinstance(res, validators.ValidationError):
        return False

    return res


def is_ipv4(v: str) -> bool:
    return _is_x(v, validator=validators.ipv4)  # type: ignore


def is_ipv6(v: str) -> bool:
    return _is_x(v, validator=validators.ipv6)  # type: ignore


def is_domain(v: str) -> bool:
    if len(v.split(".")) == 1:
        return False

    if "@" in v:
        return False

    return _is_x(v, validator=validators.domain)  # type: ignore


def is_url(v: str) -> bool:
    if not v.startswith(("http://", "https://")):
        return False

    return _is_x(v, validator=validators.url)  # type: ignore


def is_email(v: str) -> bool:
    return _is_x(v, validator=validators.email)  # type: ignore


def get_registered_domain(v: str) -> str | None:
    parsed = tldextract.extract(v)

    if parsed.registered_domain == "":
        return None

    return parsed.registered_domain


@contextmanager
def with_socket_timeout(timeout: float):
    old = socket.getdefaulttimeout()
    try:
        socket.setdefaulttimeout(timeout)
        yield
    except (socket.timeout, ValueError) as e:
        raise asyncio.TimeoutError(
            f"{timeout} seconds have passed but there is no response"
        ) from e
    finally:
        socket.setdefaulttimeout(old)


@cached(cache=TTLCache(maxsize=settings.QUERY_CACHE_SIZE, ttl=settings.QUERY_CACHE_TTL))
def _resolve(hostname: str, *, timeout: float = float(settings.QUERY_TIMEOUT)) -> str:
    with with_socket_timeout(timeout):
        return socket.gethostbyname(hostname)


resolve = asyncify(_resolve)


def load_yaml(path: str | pathlib.Path) -> dict:
    with open(path) as f:
        return cast(dict, yaml.safe_load(f))


def glob_rules(
    base_directory: str | pathlib.Path,
    *,
    additional_directories: list[str] | list[pathlib.Path] | CommaSeparatedStrings,
    rule_extensions=settings.RULE_EXTENSIONS,
) -> list[pathlib.Path]:
    directories = [base_directory]
    directories.extend(additional_directories)

    directories = [pathlib.Path(d) for d in directories]

    paths: set[str] = set()
    for directory in directories:
        for extension in rule_extensions:
            paths.update([str(p) for p in directory.glob(f"*.{extension}")])  # type: ignore

    return [pathlib.Path(p) for p in paths]



from abuse_whois import schemas
from abuse_whois.utils import is_email

from .rule import WhoisRule
from .rules import load_rules


def get_registrar_contact(record: schemas.WhoisRecord) -> schemas.Contact | None:
    provider = record.registrar
    email: str | None = None

    # check email format for just in case
    if is_email(record.abuse.email or ""):
        email = record.abuse.email

    if email is None:
        return None

    # use email's domain as a provider if provider is None
    if provider is None and is_email(email or ""):
        provider = email.split("@")[-1]

    if provider is None or email is None:
        return None

    return schemas.Contact(provider=provider, address=email)


def get_whois_contact(
    record: schemas.WhoisRecord, *, rules: list[WhoisRule] | None = None
) -> schemas.Contact | None:
    rules = rules or load_rules()
    for rule in rules:
        if rule.match(record):
            return rule.contact

    return get_registrar_contact(record)



from functools import partial
from typing import Any

import httpx
import orjson
from asyncwhois.errors import NotFoundError as WhoIsNotFoundError
from asyncwhois.pywhois import DomainLookup, NumberLookup
from httpx._exceptions import TimeoutException
from returns.functions import raise_exception
from returns.future import FutureResultE, future_safe
from returns.pipeline import flow
from returns.pointfree import bind
from whodap.errors import (
    NotFoundError as WhodapNotFoundError,
)
from whodap.errors import (
    RateLimitError,
    WhodapError,
)

from abuse_whois import errors, schemas, settings
from abuse_whois.utils import (
    get_registered_domain,
    is_domain,
    is_ipv4,
    is_ipv6,
)

from .abstract import AbstractService


def check_rate_limit(lookup: DomainLookup) -> None:
    for message in settings.WHOIS_RATE_LIMIT_MESSAGES:
        if message in lookup.query_output:
            # use whodap's RateLimitError just for convenience
            raise RateLimitError()


async def domain_query(
    address: str, *, timeout: int = settings.QUERY_TIMEOUT
) -> DomainLookup:
    async with httpx.AsyncClient(timeout=timeout) as client:
        try:
            return await DomainLookup.aio_rdap_domain(address, client)
        except (TimeoutException, WhodapError):
            # fallback to whois
            pass
        except (TypeError, NotImplementedError):
            pass

    # fallback to whois
    lookup = await DomainLookup.aio_whois_domain(
        address,
        timeout=timeout,
        authoritative_only=False,
        proxy_url=None,  # type: ignore
    )

    check_rate_limit(lookup)

    return lookup


@future_safe
async def query(
    address: str, *, timeout: int = settings.QUERY_TIMEOUT
) -> DomainLookup | NumberLookup:
    try:
        if is_domain(address):
            return await domain_query(address, timeout=timeout)

        if is_ipv4(address):
            return await NumberLookup.aio_whois_ipv4(
                address,
                timeout=timeout,
                authoritative_only=False,
                proxy_url=None,  # type: ignore
            )

        if is_ipv6(address):
            return await NumberLookup.aio_whois_ipv6(
                address,
                timeout=timeout,
                authoritative_only=False,
                proxy_url=None,  # type: ignore
            )
    except (WhodapNotFoundError, WhoIsNotFoundError) as e:
        raise errors.NotFoundError(f"Record for {address} is not found") from e
    except RateLimitError as e:
        raise errors.RateLimitError(f"Query for {address} is rate limited") from e

    raise errors.AddressError(f"{address} is neither of domain, IPv4 and IPv6")


def get_contact(parsed: dict, prefix: str) -> schemas.WhoisContact:
    name = parsed.get(f"{prefix}_name", None)
    email = parsed.get(f"{prefix}_email", None)
    telephone = parsed.get(f"{prefix}_phone", None)
    organization = parsed.get(f"{prefix}_organization", None)
    return schemas.WhoisContact(
        organization=organization, email=email, telephone=telephone, name=name
    )


def get_abuse(parsed: dict) -> schemas.WhoisAbuse:
    email = parsed.get("registrar_abuse_email", None)
    telephone = parsed.get("registrar_abuse_phone", None)
    return schemas.WhoisAbuse(
        email=email,
        telephone=telephone,
    )


@future_safe
async def normalize(hostname: str) -> str:
    if is_domain(hostname):
        return get_registered_domain(hostname) or hostname

    return hostname


def normalize_domain(domain: str | None) -> str | None:
    if domain is None:
        return None

    return domain.lower().removesuffix(".")


def is_str_list(values: Any) -> bool:
    if not isinstance(values, list):
        return False

    return all(isinstance(v, str) for v in values)


@future_safe
async def parse(result: DomainLookup | NumberLookup) -> schemas.WhoisRecord:
    parser_output = result.parser_output
    query_output = result.query_output

    raw_text: str = ""
    if isinstance(query_output, dict):
        raw_text = orjson.dumps(query_output).decode()
    else:
        raw_text = str(query_output)

    domain = normalize_domain(parser_output.get("domain_name"))

    name_servers = parser_output.get("name_servers", [])
    if not is_str_list(name_servers):
        name_servers = []

    statuses = parser_output.get("status", [])
    if not is_str_list(statuses):
        statuses = []

    return schemas.WhoisRecord(
        raw_text=raw_text,
        domain=domain,
        name_servers=name_servers,
        statuses=statuses,
        tech=get_contact(parser_output, "technical"),
        admin=get_contact(parser_output, "admin"),
        registrant=get_contact(parser_output, "registrant"),
        abuse=get_abuse(parser_output),
        expires_at=parser_output.get("expires"),
        updated_at=parser_output.get("updated"),
        registered_at=parser_output.get("registered"),
        registrar=parser_output.get("registrar"),
    )


class WhoisQuery(AbstractService):
    async def call(
        self, hostname: str, *, timeout: int = settings.QUERY_TIMEOUT
    ) -> schemas.WhoisRecord:
        result: FutureResultE[schemas.WhoisRecord] = flow(
            hostname, normalize, bind(partial(query, timeout=timeout)), bind(parse)
        )
        return (await result.awaitable()).alt(raise_exception).unwrap()._inner_value



from abuse_whois.schemas import Contact

from .rule import SharedHostingRule
from .rules import load_rules


def get_shared_hosting_provider(
    hostname: str, *, rules: list[SharedHostingRule] | None = None
) -> Contact | None:
    rules = rules or load_rules()

    for rule in rules:
        if rule.match(hostname):
            return rule.contact

    return None
